do
	local DistanceBetween = import('/lua/utilities.lua').GetDistanceBetweenTwoEntities
	local oldProjectile = Projectile
	Projectile = Class(oldProjectile)
	{	
		OnCreate = function(self, inWater)
			oldProjectile.OnCreate(self, inWater)
			
			AddOKCFunctions(self)
		end,

		--	have to do this to get Cybran torpedos working.
		--	they use a custom weapon:CreateProjectileForWeapon functon that doesn't hook.
		PassDamageData = function(self, damageData)
			oldProjectile.PassDamageData(self, damageData)
			
			if damageData.OKCData then
				self:PassOKCData(damageData.OKCData)	--	yes; we're passing this to ourself.
			end
			
			damageData.OKCData = nil	--	delete the old data, we copied it.
		end,

		PassOKCData = function(self, data)	--	map everything passed in 'data' to a local 'OKCData' table.

			self.OKCData	= data
			self.target		= data.firedAt
			
			if data.firedAt and not data.firedAt.hasOKC then
				Perf('no OKC')
				LOG('PassOKCData')
				LOG(repr(data.firedAt))
			end
			
			local tracking = self:GetBlueprint().Physics.TrackTarget 			
			if tracking and data.weapon:ShouldRetarget() and self.Trash then	--	for some reason I'm seeing errors accessing the trash bag.
				self:ForkThread(self.RetargetThread)
			elseif self.target then
				self.target:AddDamageInFlight(self.DamageData.DamageAmount or 0, tracking or data.shortLife)
			end
			
		end,
		
		OnLostTarget = function(self)	--	unsure when this would ever occur.
		
		--	LOG('*BM: Projectile lost a target.')
			
			if self:GetBlueprint().Physics.TrackTarget then
				self.Retargetting = self:ForkThread(self.RetargetThread)
			else
				oldProjectile.OnLostTarget(self)
			end
		end,
		
		--	this function must always be forked off; it may need to wait one tick to fetch data from the weapon.
		RetargetThread = function(self)
		
		--	Perf('projectile')
			
			if self.OKCData.priorities and not self.OKCData.dontOKCheck then
			
				local targetCats = ParseEntityCategory(self.OKCData.priorities[1])
				
				for id, cat in self.OKCData.priorities do
					if cat ~= 'ALLUNITS' then
						targetCats = targetCats + ParseEntityCategory(cat)
					end
				end
								
				local targetList = ArmyBrains[self:GetArmy()]:GetUnitsAroundPoint(targetCats or ParseEntityCategory('ALLUNITS'), self.OKCData.firedPos, self.OKCData.range, 'Enemy')
	
				for id, target in targetList do
					if target:BeenDestroyed() then
						continue
					end
										
					if self.OKCData.weapon:ShouldTarget(target) then
						self:TrackTarget(false)
						self:SetNewTarget(target)
						self:TrackTarget(true)
						
						if self.OKCData.firedAt == self:GetTrackingTarget() then
						--	Perf('redirect fail')
						else
						--	Perf('redirect pass')
							
							self.target = target

							self.target:AddDamageInFlight(self.DamageData.DamageAmount or 0, true)
							
							break
						end
						
					end
				end
			end
		
		end,
		
		OnImpact = function(self, targetType, targetEntity)
		
			if self.DamageData.DamageAmount and self.target and not self.OKCData.dontOKSum then
				if not self.target.hasOKC then
				--	Perf('no OKC')
					LOG('OnImpact')
					LOG(repr(self.target))
				end
			
				self.target:SubDamageInFlight(self.DamageData.DamageAmount, self:GetBlueprint().Physics.TrackTarget or self.OKCData.shortLife)
			end

			oldProjectile.OnImpact(self, targetType, targetEntity)
			
		end,
	}

end