do
	local DistanceBetweenEntities = import('/lua/utilities.lua').GetDistanceBetweenTwoEntities

	local oldWeapon = Weapon
	Weapon = Class(oldWeapon)
	{
		OnCreate = function(self)
			local bp = self:GetBlueprint()
			
		--	WARN("moho.CAiAttackerImpl_methods")
		--	WARN(repr(moho.CAiAttackerImpl_methods))
			
			--	table to hold a cache of targets for the Target Memory.
			--self.targetCache = {}
			
			self.OKC = {
				dontOKCheck			= bp.dontOKCheck,
				dontOKSum			= bp.dontOKSum,
				dontCacheTargets	= bp.dontCacheTargets
			}
			
			self.maxRadius = bp.MaxRadius or 0
			
			oldWeapon.OnCreate(self)
		end,
		
		GetCurrentTargetP = function(self)
			
			local c = self:GetCurrentTarget()
			
			if c.hasOKC then
				return c
			elseif c.GetSource then
				return c:GetSource()
			elseif not c then
				return nil	--	nothing to shoot at!
			else
				Perf('GetCurrentTarget is broken!')
				LOG(repr(c))
			end
		
		end,
		
		GetDamageTable = function(self)
			local damageTable = oldWeapon.GetDamageTable(self)
			
			if not self.OKCPass then	--	signifies that CreateProjectileForWeapon didn't hook.
				damageTable.OKCData = self:GetOKCTable()
			end
						
			return damageTable
		end,	--	GetDamageTable
		
		GetOKCTable = function(self)
			local OKCTable	= {}
			local bp		= self:GetBlueprint()
			
			OKCTable.dontOKSum	= self.OKC.dontOKSum
			OKCTable.weapon		= self
			OKCTable.firedAt	= self:GetCurrentTargetP()
			
			OKCTable.firedPos	= self.unit:GetPosition()
			
			OKCTable.range		= self.maxRadius
			OKCTable.priorities	= bp.TargetPriorities
			
			local pos = self:GetCurrentTargetPos()
			if pos then
				OKCTable.shortLife = VDist3(OKCTable.firedPos, pos) < (bp.MuzzleVelocity or 0) * 0.5	--	a short life is anything less than 0.5 seconds.
			end
			
			--if not self:GetCurrentTarget().hasOKC then
			----	Perf('no OKC')
				--LOG('GetOKCTable')
				--LOG(repr(self:GetCurrentTarget()))
				--LOG(repr(self:GetCurrentTarget():GetSource()))
			--end
			
			return OKCTable
		end,	--	GetOKCTable
		
		CreateProjectileForWeapon = function(self, bone)
			self.OKCPass = true	--	create a flag to check if CreateProjectileForWeapon hasn't been hooked properly.
								--	as in the case of the Cybran torpedo launchers.
			local proj = oldWeapon.CreateProjectileForWeapon(self, bone)
			
			proj:PassOKCData(self:GetOKCTable())

			return proj
		end,	--	CreateProjectileForWeapon		
		
		--	TARGETTING.
	
		--	guess what;
		--	you can hook, or overload this function 
		--		with something specific to your weapon.
		--	have this function return true if you 
		--		want to choose a new target.
		--	it's the weapon's job to decide if the current 
		--		target isn't any good; *not* the target's. 
		ShouldTarget = function(self, target)
			
			if not target then
				return false
			end
			
			if not target.hasOKC then
			--	Perf('no OKC')
				LOG('ShouldTarget')
				LOG(repr(target))
			end
			
		--	Perf('targetting check')
			
			return target:GetBlueprint().Defense.dontOKControl or not target:WillDie()
			
		end,	--	ShouldTarget
		
		ShouldRetarget = function(self)
			return not self:ShouldTarget(self:GetCurrentTargetP())
		end,
		
		ChangeMaxRadius = function(self, r)
			self.maxRadius = r
			oldWeapon.ChangeMaxRadius(self, r)
		end,
		
		HasCacheTargets = function(self)
		
			for k, ent in self.unit.cacheTargets do
				if ent.Dead or ent:BeenDestroyed() then
				--	LOG('removing from cache.')
				
					table.remove(self.unit.cacheTargets, k)
				end
			end
			
			--LOG(repr(self.unit.cacheTargets))
			
			return table.getn(self.unit.cacheTargets) > 0 
			
		end,
		
		IsTargetCached = function(self, target)
		
			for k, ent in self.unit.cacheTargets do
				if ent.Dead or ent:BeenDestroyed() then
				--	LOG('removing from cache.')
				
					table.remove(self.unit.cacheTargets, k)
				end
			end
		
			return table.find(self.unit.cacheTargets, target) and true
			
		end,
		
		RetargetThread = function(self)
			--	must the target be pwnt?
			--	yes:
			--		pwn it.
			--
			--	will the target get pwnt?
			--	no:
			--		is this target cached too?
			--		yes:
			--			pwn it.
			--			
			--	are there any cached targets? (at all)
			--	yes:
			--		get all nearby cached targets.
			--	no:
			--		get all nearby filter targets.
			--
			--	find a target for pwning.
			--		pwn it.
			
			local currentTarget	= self:GetCurrentTargetP()
			
			if (not currentTarget) or currentTarget:BeenDestroyed() then	--	build arms use Weapon code, and trip OnWeaponFired but don't have a target.
				return	--	whoops.
			end
					
			WaitTicks(1)
			
		--	LOG('*BM: RETARGET THREAD')
			
		--	LOG('starting ent. id is ' .. currentTarget:GetEntityId())
			
			if self.unit.Dead or self:IsTargetCached(currentTarget) and not self:ShouldRetarget() then
			--	LOG('*BM:\tstill shooting at cached target.')
			
				return	--	pwn it.
			end
			
			local bp = self:GetBlueprint()
			local targetList = {}

			if self:HasCacheTargets() and not self.OKC.dontCacheTargets then
				
			--	LOG('*BM:\tchecking the cache.')
				
				local unitPos = self.unit:GetPosition()
				
				for id, entity in self.unit.cacheTargets do
					if (not entity:BeenDestroyed()) and VDist3(unitPos, entity:GetPosition()) < self.maxRadius then
						table.insert(targetList, entity)
					end
				end
				
			--	LOG('*BM:\t\tdone. (got ' .. table.getn(targetList) .. ' things)')
				
			else
			
				if not self:ShouldRetarget() then
				
				--	LOG('*BM:\tjust going to keep shooting.')
				
					return	--	pwn it.
				end
			
				if bp.TargetPriorities then	
				
				--	Perf('target priorities')
					
				--	LOG('*BM:\tfiltering nearby units.')
					
					--	manually prime the targetCats variable so we can add to it later.
					local targetCats = ParseEntityCategory(bp.TargetPriorities[1])
					
					--	build a list of target categories to filter out some suitable targets.
					for id, cat in bp.TargetPriorities do
						if cat ~= 'ALLUNITS' then
							targetCats = targetCats + ParseEntityCategory(cat)
						end
					end
					
					targetList = ArmyBrains[self.unit:GetArmy()]:GetUnitsAroundPoint(targetCats or ParseEntityCategory('ALLUNITS'), self.unit:GetPosition(), self.maxRadius * bp.TrackingRadius, 'Enemy')
					
				--	LOG('*BM:\t\tdone.')
					
				else
				--	Perf('no priorities')
				end
				
			end
			
		--	LOG('*BM:\tchecking targets.')
			
			for id, target in targetList do
			
			--	LOG('*BM:\t\tevaluating target ' .. target:GetEntityId())
			
				if target:BeenDestroyed() then
				
				--	LOG('*BM:\t\t\thas been destroyed.')
				
					continue
				end
				
				if self:ShouldTarget(target) then
				
				--	LOG('*BM:\t\t\tit`s good to shoot at!')
				
					self:SetTargetEntity(target)
					
				--	LOG('*BM:\t\t\tset as target.')
					
					if currentTarget == self:GetCurrentTarget() then
					--	Perf('switch fail')
					else
					--	Perf('switch pass')
						
						break
					end
				
				else
				
				--	LOG('*BM:\t\t\tnot going to shoot at that.')
				
				end
			end
			
		--	LOG('*BM:\tdone retargetting.')
			
		end,	--	Retarget
		
		OnGotTarget = function(self)	--	I think this event triggers with manually issued commands. 
			oldWeapon.OnGotTarget(self)
			
		--	Perf('got target')
			
			self:ForkThread(self.RetargetThread)
		end,
	}	--	Weapon	

end